#ifndef __patchbmp_TileMap_h__
#define __patchbmp_TileMap_h__



#include <stdlib.h>
#include <string>


template <typename T>
class Buffer
{
public:
   Buffer() : mp(NULL), mn(0) {}
   ~Buffer() {delete[] mp;}
   
   Buffer(const Buffer& b) {copy(b);}
   Buffer& operator = (const Buffer& b) {delete[] mp; copy(b); return *this;}
   Buffer& attach(T* p, size_t n) {delete[] mp; mp = p; mn = n; return *this;}
   T* alloc(size_t n) {delete[] mp; mp = new T[mn=n]; return mp;}
   void release() {delete[] mp; mp = NULL; mn = 0;}

   size_t size() const {return mn;}
   operator T* () const {return mp;}
   //T& operator [] (size_t i) const {return mp[i];}
   T* operator -> () const {return mp;}


protected:
   T* mp;
   size_t mn;

   void copy(const Buffer& b) {mp = new T[mn=b.mn]; memcpy(mp, b.mp, mn);}
};


struct TileSpecs
{
   long dx0, dx1, w, dx2;
   long dy0, dy1, h, dy2;
   
   TileSpecs() : dx0(0), dx1(0), w(1), dx2(0), dy0(0), dy1(0), h(1), dy2(0) {}
   
   long blockWidth() const {return (dx1 + w + dx2);}
   long blockHeight() const {return (dy1 + h + dy2);}
};


class TileMap
{
public:
   typedef std::string string;
   
   TileMap(const string& sName);

   bool loaded() const {return mLoaded;}
   void load(const string& sFileName);
   void unload();
   
   void setTileSpecs(const TileSpecs& ts);

   void saveAs(const string& sFileName) const;

   enum Operation {COPY=0, AND, OR, XOR, nOperations};
   void mergePixelRange(Operation op, long lDestX, long lDestY,
                       const TileMap& tmSrc, long lSrcX, long lSrcY,
                       long lWidth=1, long lHeight=1);
   void mergeTileRange(Operation op, int iDestCol, int iDestRow,
                      const TileMap& tmSrc, int iSrcCol, int iSrcRow,
                      int iCols=1, int iRows=1);


protected:
   typedef unsigned char byte;
   typedef unsigned short int word;
   typedef unsigned long int dword;

   const string msName;
   bool mLoaded;
   
   TileSpecs mts;
   Buffer<byte> mpFileData; long mlFileSize;
   long mlBmpOffset;
   int miBytesPP;
   long mlWidth, mlHeight, mlLineSize;
   
   void raiseError(const string& s) const;
   void assertLoaded() const;
   
   void checkPixelCoords(long x, long y) const;
   void checkPixelRange(long x, long y, long w, long h) const;
   
   void calcTileCoords(int iCol, int iRow, long* px, long* py) const;
   long calcPixelOffset(long x, long y) const;

   void assertCompatibleWith(const TileMap& tm, bool tiles) const;

   typedef void* MergeFunc(void* pDest, const void* pSrc, size_t n);
   static void* memAND(void* pDest, const void* pSrc, size_t n);
   static void* memOR(void* pDest, const void* pSrc, size_t n);
   static void* memXOR(void* pDest, const void* pSrc, size_t n);
   static MergeFunc* const mpMergeFunc[];
};



#endif

